@@ -14,10 +14,6 @@
 
 from bisect import bisect_right
 from collections import defaultdict
-try:
-    from collections import OrderedDict
-except ImportError:  # Python <2.7
-    from cassandra.util import OrderedDict # NOQA
 from hashlib import md5
 from itertools import islice, cycle
 import json
@@ -25,16 +21,18 @@ import logging
 import re
 from threading import RLock
 import weakref
+import six
 
 murmur3 = None
 try:
-    from murmur3 import murmur3
-except ImportError:
+    from cassandra.murmur3 import murmur3
+except ImportError as e:
     pass
 
 import cassandra.cqltypes as types
 from cassandra.marshal import varint_unpack
 from cassandra.pool import Host
+from cassandra.util import OrderedDict
 
 log = logging.getLogger(__name__)
 
@@ -181,11 +179,11 @@ class Metadata(object):
 
     def _keyspace_added(self, ksname):
         if self.token_map:
-            self.token_map.rebuild_keyspace(ksname)
+            self.token_map.rebuild_keyspace(ksname, build_if_absent=False)
 
     def _keyspace_updated(self, ksname):
         if self.token_map:
-            self.token_map.rebuild_keyspace(ksname)
+            self.token_map.rebuild_keyspace(ksname, build_if_absent=False)
 
     def _keyspace_removed(self, ksname):
         if self.token_map:
@@ -333,7 +331,7 @@ class Metadata(object):
 
         token_to_host_owner = {}
         ring = []
-        for host, token_strings in token_map.iteritems():
+        for host, token_strings in six.iteritems(token_map):
             for token_string in token_strings:
                 token = token_class(token_string)
                 ring.append(token)
@@ -462,7 +460,8 @@ class NetworkTopologyStrategy(ReplicationStrategy):
     """
 
     def __init__(self, dc_replication_factors):
-        self.dc_replication_factors = dc_replication_factors
+        self.dc_replication_factors = dict(
+                (str(k), int(v)) for k, v in dc_replication_factors.items())
 
     def make_token_replica_map(self, token_to_host_owner, ring):
         # note: this does not account for hosts having different racks
@@ -796,14 +795,18 @@ class TableMetadata(object):
         return list(sorted(ret))
 
 
-def protect_name(name):
-    if isinstance(name, unicode):
-        name = name.encode('utf8')
-    return maybe_escape_name(name)
+if six.PY3:
+    def protect_name(name):
+        return maybe_escape_name(name)
+else:
+    def protect_name(name):  # NOQA
+        if isinstance(name, six.text_type):
+            name = name.encode('utf8')
+        return maybe_escape_name(name)
 
 
 def protect_names(names):
-    return map(protect_name, names)
+    return [protect_name(n) for n in names]
 
 
 def protect_value(value):
@@ -950,10 +953,14 @@ class TokenMap(object):
 
         self.tokens_to_hosts_by_ks = {}
         self._metadata = metadata
+        self._rebuild_lock = RLock()
 
-    def rebuild_keyspace(self, keyspace):
-        self.tokens_to_hosts_by_ks[keyspace] = \
-            self.replica_map_for_keyspace(self._metadata.keyspaces[keyspace])
+    def rebuild_keyspace(self, keyspace, build_if_absent=False):
+        with self._rebuild_lock:
+            current = self.tokens_to_hosts_by_ks.get(keyspace, None)
+            if (build_if_absent and current is None) or (not build_if_absent and current is not None):
+                replica_map = self.replica_map_for_keyspace(self._metadata.keyspaces[keyspace])
+                self.tokens_to_hosts_by_ks[keyspace] = replica_map
 
     def replica_map_for_keyspace(self, ks_metadata):
         strategy = ks_metadata.replication_strategy
@@ -972,7 +979,7 @@ class TokenMap(object):
         """
         tokens_to_hosts = self.tokens_to_hosts_by_ks.get(keyspace, None)
         if tokens_to_hosts is None:
-            self.rebuild_keyspace(keyspace)
+            self.rebuild_keyspace(keyspace, build_if_absent=True)
             tokens_to_hosts = self.tokens_to_hosts_by_ks.get(keyspace, None)
             if tokens_to_hosts is None:
                 return []
@@ -1011,11 +1018,14 @@ class Token(object):
     def __eq__(self, other):
         return self.value == other.value
 
+    def __lt__(self, other):
+        return self.value < other.value
+
     def __hash__(self):
         return hash(self.value)
 
     def __repr__(self):
-        return "<%s: %r>" % (self.__class__.__name__, self.value)
+        return "<%s: %s>" % (self.__class__.__name__, self.value)
     __str__ = __repr__
 
 MIN_LONG = -(2 ** 63)
@@ -1034,7 +1044,7 @@ class Murmur3Token(Token):
     @classmethod
     def hash_fn(cls, key):
         if murmur3 is not None:
-            h = murmur3(key)
+            h = int(murmur3(key))
             return h if h != MIN_LONG else MAX_LONG
         else:
             raise NoMurmur3()
@@ -1051,6 +1061,8 @@ class MD5Token(Token):
 
     @classmethod
     def hash_fn(cls, key):
+        if isinstance(key, six.text_type):
+            key = key.encode('UTF-8')
         return abs(varint_unpack(md5(key).digest()))
 
     def __init__(self, token):
@@ -1065,7 +1077,7 @@ class BytesToken(Token):
 
     def __init__(self, token_string):
         """ `token_string` should be string representing the token. """
-        if not isinstance(token_string, basestring):
+        if not isinstance(token_string, six.string_types):
             raise TypeError(
                 "Tokens for ByteOrderedPartitioner should be strings (got %s)"
                 % (type(token_string),))
