@@ -16,9 +16,12 @@ from itertools import islice, cycle, groupby, repeat
 import logging
 from random import randint
 from threading import Lock
+import six
 
 from cassandra import ConsistencyLevel
 
+from six.moves import range
+
 log = logging.getLogger(__name__)
 
 
@@ -263,7 +266,7 @@ class DCAwareRoundRobinPolicy(LoadBalancingPolicy):
         for host in islice(cycle(local_live), pos, pos + len(local_live)):
             yield host
 
-        for dc, current_dc_hosts in self._dc_live_hosts.iteritems():
+        for dc, current_dc_hosts in six.iteritems(self._dc_live_hosts):
             if dc == self.local_dc:
                 continue
 
@@ -391,7 +394,8 @@ class WhiteListRoundRobinPolicy(RoundRobinPolicy):
     """
     def __init__(self, hosts):
         """
-        :param hosts: List of hosts
+        The `hosts` parameter should be a sequence of hosts to permit
+        connections to.
         """
         self._allowed_hosts = hosts
         RoundRobinPolicy.__init__(self)
@@ -529,7 +533,7 @@ class ExponentialReconnectionPolicy(ReconnectionPolicy):
         self.max_delay = max_delay
 
     def new_schedule(self):
-        return (min(self.base_delay * (2 ** i), self.max_delay) for i in xrange(64))
+        return (min(self.base_delay * (2 ** i), self.max_delay) for i in range(64))
 
 
 class WriteType(object):
@@ -568,6 +572,14 @@ class WriteType(object):
     internally before a BATCH write.
     """
 
+WriteType.name_to_value = {
+    'SIMPLE': WriteType.SIMPLE,
+    'BATCH': WriteType.BATCH,
+    'UNLOGGED_BATCH': WriteType.UNLOGGED_BATCH,
+    'COUNTER': WriteType.COUNTER,
+    'BATCH_LOG': WriteType.BATCH_LOG,
+}
+
 
 class RetryPolicy(object):
     """
