@@ -18,24 +18,126 @@ specifying consistency levels and retry policies for individual
 queries.
 """
 
+from collections import namedtuple
 from datetime import datetime, timedelta
+import re
 import struct
 import time
+import six
 
 from cassandra import ConsistencyLevel, OperationTimedOut
 from cassandra.cqltypes import unix_time_from_uuid1
-from cassandra.decoder import (cql_encoders, cql_encode_object,
-                               cql_encode_sequence, named_tuple_factory)
+from cassandra.encoder import (cql_encoders, cql_encode_object,
+                               cql_encode_sequence)
+from cassandra.util import OrderedDict
 
 import logging
 log = logging.getLogger(__name__)
 
 
+NON_ALPHA_REGEX = re.compile('[^a-zA-Z0-9]')
+START_BADCHAR_REGEX = re.compile('^[^a-zA-Z0-9]*')
+END_BADCHAR_REGEX = re.compile('[^a-zA-Z0-9_]*$')
+
+_clean_name_cache = {}
+
+
+def _clean_column_name(name):
+    try:
+        return _clean_name_cache[name]
+    except KeyError:
+        clean = NON_ALPHA_REGEX.sub("_", START_BADCHAR_REGEX.sub("", END_BADCHAR_REGEX.sub("", name)))
+        _clean_name_cache[name] = clean
+        return clean
+
+
+def tuple_factory(colnames, rows):
+    """
+    Returns each row as a tuple
+
+    Example::
+
+        >>> from cassandra.query import named_tuple_factory
+        >>> session = cluster.connect('mykeyspace')
+        >>> session.row_factory = tuple_factory
+        >>> rows = session.execute("SELECT name, age FROM users LIMIT 1")
+        >>> print rows[0]
+        ('Bob', 42)
+
+    .. versionchanged:: 2.0.0
+        moved from ``cassandra.decoder`` to ``cassandra.query``
+    """
+    return rows
+
+
+def named_tuple_factory(colnames, rows):
+    """
+    Returns each row as a `namedtuple <https://docs.python.org/2/library/collections.html#collections.namedtuple>`_.
+    This is the default row factory.
+
+    Example::
+
+        >>> from cassandra.query import named_tuple_factory
+        >>> session = cluster.connect('mykeyspace')
+        >>> session.row_factory = named_tuple_factory
+        >>> rows = session.execute("SELECT name, age FROM users LIMIT 1")
+        >>> user = rows[0]
+
+        >>> # you can access field by their name:
+        >>> print "name: %s, age: %d" % (user.name, user.age)
+        name: Bob, age: 42
+
+        >>> # or you can access fields by their position (like a tuple)
+        >>> name, age = user
+        >>> print "name: %s, age: %d" % (name, age)
+        name: Bob, age: 42
+        >>> name = user[0]
+        >>> age = user[1]
+        >>> print "name: %s, age: %d" % (name, age)
+        name: Bob, age: 42
+
+    .. versionchanged:: 2.0.0
+        moved from ``cassandra.decoder`` to ``cassandra.query``
+    """
+    Row = namedtuple('Row', map(_clean_column_name, colnames))
+    return [Row(*row) for row in rows]
+
+
+def dict_factory(colnames, rows):
+    """
+    Returns each row as a dict.
+
+    Example::
+
+        >>> from cassandra.query import named_tuple_factory
+        >>> session = cluster.connect('mykeyspace')
+        >>> session.row_factory = dict_factory
+        >>> rows = session.execute("SELECT name, age FROM users LIMIT 1")
+        >>> print rows[0]
+        {'age': 42, 'name': 'Bob'}
+
+    .. versionchanged:: 2.0.0
+        moved from ``cassandra.decoder`` to ``cassandra.query``
+    """
+    return [dict(zip(colnames, row)) for row in rows]
+
+
+def ordered_dict_factory(colnames, rows):
+    """
+    Like :meth:`~cassandra.query.dict_factory`, but returns each row as an OrderedDict,
+    so the order of the columns is preserved.
+
+    .. versionchanged:: 2.0.0
+        moved from ``cassandra.decoder`` to ``cassandra.query``
+    """
+    return [OrderedDict(zip(colnames, row)) for row in rows]
+
+
 class Statement(object):
     """
-    An abstract class representing a single query. There are two subclasses:
-    :class:`.SimpleStatement` and :class:`.BoundStatement`.  These can
-    be passed to :meth:`.Session.execute()`.
+    An abstract class representing a single query. There are three subclasses:
+    :class:`.SimpleStatement`, :class:`.BoundStatement`, and :class:`.BatchStatement`.
+    These can be passed to :meth:`.Session.execute()`.
     """
 
     retry_policy = None
@@ -51,18 +153,36 @@ class Statement(object):
     this will be set to a :class:`.QueryTrace` instance.
     """
 
-    consistency_level = ConsistencyLevel.ONE
+    consistency_level = None
     """
     The :class:`.ConsistencyLevel` to be used for this operation.  Defaults
-    to :attr:`.ConsistencyLevel.ONE`.
+    to :const:`None`, which means that the default consistency level for
+    the Session this is executed in will be used.
     """
 
+    fetch_size = None
+    """
+    How many rows will be fetched at a time.  This overrides the default
+    of :attr:`.Session.default_fetch_size`
+
+    This only takes effect when protocol version 2 or higher is used.
+    See :attr:`.Cluster.protocol_version` for details.
+
+    .. versionadded:: 2.0.0
+    """
+
+    _serial_consistency_level = None
     _routing_key = None
 
-    def __init__(self, retry_policy=None, consistency_level=None, routing_key=None):
+    def __init__(self, retry_policy=None, consistency_level=None, routing_key=None,
+                 serial_consistency_level=None, fetch_size=None):
         self.retry_policy = retry_policy
         if consistency_level is not None:
             self.consistency_level = consistency_level
+        if serial_consistency_level is not None:
+            self.serial_consistency_level = serial_consistency_level
+        if fetch_size is not None:
+            self.fetch_size = None
         self._routing_key = routing_key
 
     def _get_routing_key(self):
@@ -70,8 +190,8 @@ class Statement(object):
 
     def _set_routing_key(self, key):
         if isinstance(key, (list, tuple)):
-            self._routing_key = "".join(struct.pack("HsB", len(component), component, 0)
-                                        for component in key)
+            self._routing_key = b"".join(struct.pack("HsB", len(component), component, 0)
+                                         for component in key)
         else:
             self._routing_key = key
 
@@ -91,6 +211,54 @@ class Statement(object):
         components should be strings.
         """)
 
+    def _get_serial_consistency_level(self):
+        return self._serial_consistency_level
+
+    def _set_serial_consistency_level(self, serial_consistency_level):
+        acceptable = (None, ConsistencyLevel.SERIAL, ConsistencyLevel.LOCAL_SERIAL)
+        if serial_consistency_level not in acceptable:
+            raise ValueError(
+                "serial_consistency_level must be either ConsistencyLevel.SERIAL "
+                "or ConsistencyLevel.LOCAL_SERIAL")
+
+    def _del_serial_consistency_level(self):
+        self._serial_consistency_level = None
+
+    serial_consistency_level = property(
+         _get_serial_consistency_level,
+         _set_serial_consistency_level,
+         _del_serial_consistency_level,
+        """
+        The serial consistency level is only used by conditional updates
+        (``INSERT``, ``UPDATE`` and ``DELETE`` with an ``IF`` condition).  For
+        those, the ``serial_consistency_level`` defines the consistency level of
+        the serial phase (or "paxos" phase) while the normal
+        :attr:`~.consistency_level` defines the consistency for the "learn" phase,
+        i.e. what type of reads will be guaranteed to see the update right away.
+        For example, if a conditional write has a :attr:`~.consistency_level` of
+        :attr:`~.ConsistencyLevel.QUORUM` (and is successful), then a
+        :attr:`~.ConsistencyLevel.QUORUM` read is guaranteed to see that write.
+        But if the regular :attr:`~.consistency_level` of that write is
+        :attr:`~.ConsistencyLevel.ANY`, then only a read with a
+        :attr:`~.consistency_level` of :attr:`~.ConsistencyLevel.SERIAL` is
+        guaranteed to see it (even a read with consistency
+        :attr:`~.ConsistencyLevel.ALL` is not guaranteed to be enough).
+
+        The serial consistency can only be one of :attr:`~.ConsistencyLevel.SERIAL`
+        or :attr:`~.ConsistencyLevel.LOCAL_SERIAL`. While ``SERIAL`` guarantees full
+        linearizability (with other ``SERIAL`` updates), ``LOCAL_SERIAL`` only
+        guarantees it in the local data center.
+
+        The serial consistency level is ignored for any query that is not a
+        conditional update. Serial reads should use the regular
+        :attr:`consistency_level`.
+
+        Serial consistency levels may only be used against Cassandra 2.0+
+        and the :attr:`~.Cluster.protocol_version` must be set to 2 or higher.
+
+        .. versionadded:: 2.0.0
+        """)
+
     @property
     def keyspace(self):
         """
@@ -119,7 +287,7 @@ class SimpleStatement(Statement):
         return self._query_string
 
     def __str__(self):
-        consistency = ConsistencyLevel.value_to_name[self.consistency_level]
+        consistency = ConsistencyLevel.value_to_name.get(self.consistency_level, 'Not Set')
         return (u'<SimpleStatement query="%s", consistency=%s>' %
                 (self.query_string, consistency))
     __repr__ = __str__
@@ -142,16 +310,19 @@ class PreparedStatement(object):
 
     routing_key_indexes = None
 
-    consistency_level = ConsistencyLevel.ONE
+    consistency_level = None
+    serial_consistency_level = None
 
     def __init__(self, column_metadata, query_id, routing_key_indexes, query, keyspace,
-                 consistency_level=ConsistencyLevel.ONE):
+                 consistency_level=None, serial_consistency_level=None, fetch_size=None):
         self.column_metadata = column_metadata
         self.query_id = query_id
         self.routing_key_indexes = routing_key_indexes
         self.query_string = query
         self.keyspace = keyspace
         self.consistency_level = consistency_level
+        self.serial_consistency_level = serial_consistency_level
+        self.fetch_size = fetch_size
 
     @classmethod
     def from_message(cls, query_id, column_metadata, cluster_metadata, query, keyspace):
@@ -190,7 +361,7 @@ class PreparedStatement(object):
         return BoundStatement(self).bind(values)
 
     def __str__(self):
-        consistency = ConsistencyLevel.value_to_name[self.consistency_level]
+        consistency = ConsistencyLevel.value_to_name.get(self.consistency_level, 'Not Set')
         return (u'<PreparedStatement query="%s", consistency=%s>' %
                 (self.query_string, consistency))
     __repr__ = __str__
@@ -220,6 +391,7 @@ class BoundStatement(Statement):
         All other ``*args`` and ``**kwargs`` will be passed to :class:`.Statement`.
         """
         self.consistency_level = prepared_statement.consistency_level
+        self.serial_consistency_level = prepared_statement.serial_consistency_level
         self.prepared_statement = prepared_statement
         self.values = []
 
@@ -232,6 +404,8 @@ class BoundStatement(Statement):
         * a sequence, even if you are only binding one value, or
         * a dict that relates 1-to-1 between dict keys and columns
         """
+        if values is None:
+            values = ()
         col_meta = self.prepared_statement.column_metadata
 
         # special case for binding dicts
@@ -312,7 +486,7 @@ class BoundStatement(Statement):
                 val = self.values[statement_index]
                 components.append(struct.pack("HsB", len(val), val, 0))
 
-            self._routing_key = "".join(components)
+            self._routing_key = b"".join(components)
 
         return self._routing_key
 
@@ -325,12 +499,154 @@ class BoundStatement(Statement):
             return None
 
     def __str__(self):
-        consistency = ConsistencyLevel.value_to_name[self.consistency_level]
+        consistency = ConsistencyLevel.value_to_name.get(self.consistency_level, 'Not Set')
         return (u'<BoundStatement query="%s", values=%s, consistency=%s>' %
                 (self.prepared_statement.query_string, self.raw_values, consistency))
     __repr__ = __str__
 
 
+class BatchType(object):
+    """
+    A BatchType is used with :class:`.BatchStatement` instances to control
+    the atomicity of the batch operation.
+
+    .. versionadded:: 2.0.0
+    """
+
+    LOGGED = None
+    """
+    Atomic batch operation.
+    """
+
+    UNLOGGED = None
+    """
+    Non-atomic batch operation.
+    """
+
+    COUNTER = None
+    """
+    Batches of counter operations.
+    """
+
+    def __init__(self, name, value):
+        self.name = name
+        self.value = value
+
+    def __str__(self):
+        return self.name
+
+    def __repr__(self):
+        return "BatchType.%s" % (self.name, )
+
+
+BatchType.LOGGED = BatchType("LOGGED", 0)
+BatchType.UNLOGGED = BatchType("UNLOGGED", 1)
+BatchType.COUNTER = BatchType("COUNTER", 2)
+
+
+class BatchStatement(Statement):
+    """
+    A protocol-level batch of operations which are applied atomically
+    by default.
+
+    .. versionadded:: 2.0.0
+    """
+
+    batch_type = None
+    """
+    The :class:`.BatchType` for the batch operation.  Defaults to
+    :attr:`.BatchType.LOGGED`.
+    """
+
+    _statements_and_parameters = None
+
+    def __init__(self, batch_type=BatchType.LOGGED, retry_policy=None,
+                 consistency_level=None):
+        """
+        `batch_type` specifies The :class:`.BatchType` for the batch operation.
+        Defaults to :attr:`.BatchType.LOGGED`.
+
+        `retry_policy` should be a :class:`~.RetryPolicy` instance for
+        controlling retries on the operation.
+
+        `consistency_level` should be a :class:`~.ConsistencyLevel` value
+        to be used for all operations in the batch.
+
+        Example usage:
+
+        .. code-block:: python
+
+            insert_user = session.prepare("INSERT INTO users (name, age) VALUES (?, ?)")
+            batch = BatchStatement(consistency_level=ConsistencyLevel.QUORUM)
+
+            for (name, age) in users_to_insert:
+                batch.add(insert_user, (name, age))
+
+            session.execute(batch)
+
+        You can also mix different types of operations within a batch:
+
+        .. code-block:: python
+
+            batch = BatchStatement()
+            batch.add(SimpleStatement("INSERT INTO users (name, age) VALUES (%s, %s)", (name, age))
+            batch.add(SimpleStatement("DELETE FROM pending_users WHERE name=%s", (name,))
+            session.execute(batch)
+
+        .. versionadded:: 2.0.0
+        """
+        self.batch_type = batch_type
+        self._statements_and_parameters = []
+        Statement.__init__(self, retry_policy=retry_policy, consistency_level=consistency_level)
+
+    def add(self, statement, parameters=None):
+        """
+        Adds a :class:`.Statement` and optional sequence of parameters
+        to be used with the statement to the batch.
+
+        Like with other statements, parameters must be a sequence, even
+        if there is only one item.
+        """
+        if isinstance(statement, six.string_types):
+            if parameters:
+                statement = bind_params(statement, parameters)
+            self._statements_and_parameters.append((False, statement, ()))
+        elif isinstance(statement, PreparedStatement):
+            query_id = statement.query_id
+            bound_statement = statement.bind(() if parameters is None else parameters)
+            self._statements_and_parameters.append(
+                (True, query_id, bound_statement.values))
+        elif isinstance(statement, BoundStatement):
+            if parameters:
+                raise ValueError(
+                    "Parameters cannot be passed with a BoundStatement "
+                    "to BatchStatement.add()")
+            self._statements_and_parameters.append(
+                (True, statement.prepared_statement.query_id, statement.values))
+        else:
+            # it must be a SimpleStatement
+            query_string = statement.query_string
+            if parameters:
+                query_string = bind_params(query_string, parameters)
+            self._statements_and_parameters.append((False, query_string, ()))
+        return self
+
+    def add_all(self, statements, parameters):
+        """
+        Adds a sequence of :class:`.Statement` objects and a matching sequence
+        of parameters to the batch.  :const:`None` can be used in place of
+        parameters when no parameters are needed.
+        """
+        for statement, value in zip(statements, parameters):
+            self.add(statement, parameters)
+
+    def __str__(self):
+        consistency = ConsistencyLevel.value_to_name.get(self.consistency_level, 'Not Set')
+        return (u'<BatchStatement type=%s, statements=%d, consistency=%s>' %
+                (self.batch_type, len(self._statements_and_parameters), consistency))
+    __repr__ = __str__
+
+
 class ValueSequence(object):
     """
     A wrapper class that is used to specify that a sequence of values should
@@ -355,11 +671,9 @@ class ValueSequence(object):
 
 def bind_params(query, params):
     if isinstance(params, dict):
-        return query % dict((k, cql_encoders.get(type(v), cql_encode_object)(v))
-                            for k, v in params.iteritems())
+        return query % dict((k, cql_encoders.get(type(v), cql_encode_object)(v)) for k, v in six.iteritems(params))
     else:
-        return query % tuple(cql_encoders.get(type(v), cql_encode_object)(v)
-                             for v in params)
+        return query % tuple(cql_encoders.get(type(v), cql_encode_object)(v) for v in params)
 
 
 class TraceUnavailable(Exception):
